/*
 * Copyright 2020 RationalityFrontline
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

@file:Suppress("NOTHING_TO_INLINE")

package org.rationalityfrontline.kevent

import org.rationalityfrontline.kevent.KEvent.DispatchMode
import org.rationalityfrontline.kevent.KEvent.ThreadMode

/**
 * Convenient interface for adding and removing subscribers
 */
interface KEventSubscriber {
    /**
     * The subscription tag of this object.
     */
    val SUBSCRIPTION_TAG: String get() = "${this::class.java.simpleName}@${this.hashCode()}"
}

/**
 * Add a new subscriber with tag of [SUBSCRIPTION_TAG]. Subscribers with the same [consumer] can only be added once.
 *
 * @param eventType event type that this subscriber subscribers.
 * @param consumer event consumer function.
 * @param threadMode thread mode, see [ThreadMode].
 * @param priority priority, bigger is higher.
 * @return true if subscription is successful, else false.
 */
inline fun <T : Any> KEventSubscriber.subscribe(
    eventType: Enum<*>,
    threadMode: ThreadMode = KEvent.DEFAULT_THREAD_MODE,
    priority: Int = KEvent.DEFAULT_PRIORITY,
    noinline consumer: EventConsumer<T>
): Boolean = KEvent.subscribe(eventType, threadMode, priority, SUBSCRIPTION_TAG, consumer)

/**
 * Add a new subscriber with tag of [SUBSCRIPTION_TAG]. Subscribers with the same [consumer] can only be added once.
 *
 * @param eventType event type that this subscriber subscribers.
 * @param consumer event consumer function.
 * @param threadMode thread mode, see [ThreadMode].
 * @param priority priority, bigger is higher.
 * @return true if subscription is successful, else false.
 */
inline fun <T : Any> KEventSubscriber.subscribe(
    eventType: Enum<*>,
    noinline consumer: EventConsumer<T>,
    threadMode: ThreadMode = KEvent.DEFAULT_THREAD_MODE,
    priority: Int = KEvent.DEFAULT_PRIORITY
): Boolean = KEvent.subscribe(eventType, consumer, threadMode, priority, SUBSCRIPTION_TAG)

/**
 * Add a new subscriber with multiple [eventTypes] and tag of [SUBSCRIPTION_TAG]. Subscribers with the same [consumer] can only be added once.
 *
 * @param eventTypes event types that this subscriber subscribers.
 * @param consumer event consumer function.
 * @param threadMode thread mode, see [ThreadMode].
 * @param priority priority, bigger is higher.
 * @return true if subscription is successful, else false.
 */
inline fun <T : Any> KEventSubscriber.subscribeMultiple(
    eventTypes: Collection<Enum<*>>,
    threadMode: ThreadMode = KEvent.DEFAULT_THREAD_MODE,
    priority: Int = KEvent.DEFAULT_PRIORITY,
    noinline consumer: EventConsumer<T>
): Boolean = KEvent.subscribeMultiple(eventTypes, threadMode, priority, SUBSCRIPTION_TAG, consumer)

/**
 * Add a new subscriber with multiple [eventTypes] and tag of [SUBSCRIPTION_TAG]. Subscribers with the same [consumer] can only be added once.
 *
 * @param eventTypes event types that this subscriber subscribers.
 * @param consumer event consumer function.
 * @param threadMode thread mode, see [ThreadMode].
 * @param priority priority, bigger is higher.
 * @return true if subscription is successful, else false.
 */
inline fun <T : Any> KEventSubscriber.subscribeMultiple(
    eventTypes: Collection<Enum<*>>,
    noinline consumer: EventConsumer<T>,
    threadMode: ThreadMode = KEvent.DEFAULT_THREAD_MODE,
    priority: Int = KEvent.DEFAULT_PRIORITY,
): Boolean = KEvent.subscribeMultiple(eventTypes, consumer, threadMode, priority, SUBSCRIPTION_TAG)

/**
 * Unsubscribe the subscriber with [eventType] and [consumer].
 *
 * @return true if subscriber exists and unsubscription is successful, else false.
 */
inline fun <T : Any> KEventSubscriber.unsubscribe(
    eventType: Enum<*>,
    noinline consumer: EventConsumer<T>
): Boolean = KEvent.unsubscribe(eventType, consumer)

/**
 * Unsubscribe the subscriber with [eventTypes] and [consumer].
 *
 * @return true if subscriber exists and any of the unsubscription is successful, else false.
 */
inline fun <T : Any> KEventSubscriber.unsubscribeMultiple(
    eventTypes: Collection<Enum<*>>,
    noinline consumer: EventConsumer<T>
): Boolean = KEvent.unsubscribeMultiple(eventTypes, consumer)

/**
 * Unsubscribe all subscribers with tag of [SUBSCRIPTION_TAG].
 *
 * @return true if subscriber exists and any of the unsubscription is successful, else false.
 */
inline fun KEventSubscriber.unsubscribeAll(): Boolean = KEvent.removeSubscribersByTag(SUBSCRIPTION_TAG)

/**
 * Cancel further event dispatching, this should only be used together with [DispatchMode.SEQUENTIAL] or [DispatchMode.INSTANTLY].
 *
 * @return true if the event get cancelled, false if the event is already cancelled before.
 */
inline fun <T : Any> KEventSubscriber.cancelEvent(event: Event<T>): Boolean = KEvent.cancelEvent(event)